use crate::matrix::Matrix;
use nom::{
    bytes::complete::tag,
    character::complete::{anychar, char, space0, space1, u8},
    combinator::eof,
    multi::{many0, many1, separated_list0, separated_list1},
    sequence::{delimited, terminated},
    IResult, Parser,
};
use std::fmt::Debug;

const INPUT: &str = include_str!("../data/day5.txt");

pub fn part1() -> String {
    let (stacks, commands) = input(INPUT);
    tops(run_commands(stacks, commands))
}

pub fn part2() -> String {
    let (stacks, commands) = input(INPUT);
    tops(run_commands_multicrate(stacks, commands))
}

fn run_commands(stacks: Stacks<char>, commands: Vec<Command>) -> Stacks<char> {
    commands
        .into_iter()
        .fold(stacks, |stacks, command| stacks.run_command(command))
}

fn run_commands_multicrate(stacks: Stacks<char>, commands: Vec<Command>) -> Stacks<char> {
    commands.into_iter().fold(stacks, |stacks, command| {
        stacks.run_command_multicrate(command)
    })
}

fn tops(stacks: Stacks<char>) -> String {
    stacks
        .0
        .iter()
        .map(|stack| stack.top())
        .filter_map(|v| v)
        .collect::<String>()
}

fn print_matrix(matrix: &Matrix<Option<char>>) {
    for row in matrix.data.clone() {
        for val in row {
            print!(" {} ", val.unwrap_or(' '));
        }
        println!("");
    }
}

#[derive(Clone, Debug, PartialEq)]
struct Command {
    count: u8,
    source: u8,
    dest: u8,
}

#[derive(Debug)]
struct Stack<T>(Vec<T>);

impl<T: Clone + Copy> Stack<T> {
    fn from_option_vec(data: Vec<Option<T>>) -> Self {
        Stack(data.into_iter().filter_map(|v| v).collect::<Vec<T>>())
    }

    fn push(&mut self, val: T) {
        self.0.push(val);
    }

    fn pop(&mut self) -> Option<T> {
        self.0.pop()
    }

    fn top(&self) -> Option<T> {
        if self.0.len() > 0 {
            Some(self.0[self.0.len() - 1])
        } else {
            None
        }
    }
}

#[derive(Debug)]
struct Stacks<T>(Vec<Stack<T>>);

impl<T: Clone + Copy> Stacks<T> {
    fn run_command(mut self, cmd: Command) -> Self {
        let Command {
            count,
            source,
            dest,
        } = cmd;
        (0..count).for_each(|_| {
            let val = self.0[(source - 1) as usize].pop().unwrap();
            self.0[(dest - 1) as usize].push(val);
        });
        self
    }

    fn run_command_multicrate(mut self, cmd: Command) -> Self {
        let Command {
            count,
            source,
            dest,
        } = cmd;

        let mut stack = Vec::new();
        (0..count).for_each(|_| {
            let val = self.0[(source - 1) as usize].pop().unwrap();
            stack.push(val);
        });
        while stack.len() > 0 {
            self.0[(dest - 1) as usize].push(stack.pop().unwrap());
        }

        self
    }
}

fn input(data: &str) -> (Stacks<char>, Vec<Command>) {
    let (_, result) = parser(data).unwrap();
    result
}

fn parser(input: &str) -> IResult<&str, (Stacks<char>, Vec<Command>)> {
    let (input, rows) = many0(parse_crate_row)(input)?;

    let (input, _) = parse_stack_numbers(input)?;

    let (input, _) = space0(input)?;
    let (input, _) = tag("\n")(input)?;

    let (input, commands) = many1(terminated(parse_command, tag("\n").or(eof)))(input)?;

    let matrix = Matrix::new(rows);

    let stacks = matrix
        .rotate_cw()
        .data
        .into_iter()
        .map(|v| Stack::from_option_vec(v.clone()))
        .collect::<Vec<Stack<char>>>();
    Ok((input, (Stacks(stacks), commands)))
}

fn parse_command(input: &str) -> IResult<&str, Command> {
    let (input, _) = tag("move ")(input)?;
    let (input, count) = u8(input)?;
    let (input, _) = tag(" from ")(input)?;
    let (input, source) = u8(input)?;
    let (input, _) = tag(" to ")(input)?;
    let (input, dest) = u8(input)?;

    Ok((
        input,
        Command {
            count,
            source,
            dest,
        },
    ))
}

fn parse_stack_numbers(input: &str) -> IResult<&str, usize> {
    let (input, _) = space0(input)?;
    let (input, stack_numbers) = separated_list1(space1, u8)(input)?;
    let (input, _) = space0(input)?;
    let (input, _) = tag("\n")(input)?;
    Ok((input, stack_numbers.len()))
}

fn parse_crate_row(input: &str) -> IResult<&str, Vec<Option<char>>> {
    let (input, cells) = terminated(separated_list0(char(' '), parse_cell), tag("\n"))(input)?;
    Ok((input, cells))
}

fn parse_cell(input: &str) -> IResult<&str, Option<char>> {
    parse_crate.or(parse_space).parse(input)
}

fn parse_crate(input: &str) -> IResult<&str, Option<char>> {
    delimited(char('['), anychar, char(']'))
        .map(|c| Some(c))
        .parse(input)
}

fn parse_space(input: &str) -> IResult<&str, Option<char>> {
    tag("   ").map(|_| None).parse(input)
}

#[cfg(test)]
mod test {
    use super::*;

    const TEST_DATA: &str = "    [D]    
[N] [C]    
[Z] [M] [P]
 1   2   3 

move 1 from 2 to 1
move 3 from 1 to 3
move 2 from 2 to 1
move 1 from 1 to 2";

    fn with_input<F>(test: F)
    where
        F: Fn((Stacks<char>, Vec<Command>)) -> () + std::panic::UnwindSafe,
    {
        test(input(TEST_DATA));
    }

    #[test]
    fn it_can_parse_test_data() {
        with_input(|(stacks, commands)| {
            assert_eq!(stacks.0[0].0, vec!['Z', 'N']);
            assert_eq!(stacks.0[1].0, vec!['M', 'C', 'D']);
            assert_eq!(stacks.0[2].0, vec!['P']);

            assert_eq!(commands.len(), 4);
            assert_eq!(
                commands[0],
                Command {
                    count: 1,
                    source: 2,
                    dest: 1
                }
            );
            assert_eq!(
                commands[3],
                Command {
                    count: 1,
                    source: 1,
                    dest: 2
                }
            );
        });
    }

    #[test]
    fn it_can_follow_commands() {
        with_input(|(stacks, commands)| {
            let stacks = stacks.run_command(commands[0].clone());
            assert_eq!(stacks.0[0].0, vec!['Z', 'N', 'D']);
            assert_eq!(stacks.0[1].0, vec!['M', 'C']);
            assert_eq!(stacks.0[2].0, vec!['P']);
        });

        with_input(|(stacks, commands)| {
            let stacks = run_commands(stacks, commands);
            assert_eq!(stacks.0[0].0, vec!['C']);
            assert_eq!(stacks.0[1].0, vec!['M']);
            assert_eq!(stacks.0[2].0, vec!['P', 'D', 'N', 'Z']);
        });
    }

    #[test]
    fn it_can_follow_multicrate_commands() {
        with_input(|(stacks, commands)| {
            let stacks = stacks.run_command_multicrate(commands[0].clone());
            assert_eq!(stacks.0[0].0, vec!['Z', 'N', 'D']);
            assert_eq!(stacks.0[1].0, vec!['M', 'C']);
            assert_eq!(stacks.0[2].0, vec!['P']);
        });

        with_input(|(stacks, commands)| {
            let stacks = run_commands_multicrate(stacks, commands);
            assert_eq!(stacks.0[0].0, vec!['M']);
            assert_eq!(stacks.0[1].0, vec!['C']);
            assert_eq!(stacks.0[2].0, vec!['P', 'Z', 'N', 'D']);
        });
    }

    #[test]
    fn it_can_find_tops() {
        with_input(|(stacks, commands)| {
            let stacks = run_commands(stacks, commands);
            assert_eq!(tops(stacks), "CMZ");
        });
    }
}
