use std::collections::{HashMap, HashSet};

const INPUT: &str = include_str!("../data/day4.txt");

pub fn day4a() -> String {
    format!(
        "{}",
        parse_cards(INPUT)
            .map(|card| card.points())
            .fold(0, |acc, points| acc + points)
    )
}

pub fn day4b() -> String {
    format!("{}", count_cards(INPUT))
}

fn parse_cards<'a>(s: &'a str) -> impl Iterator<Item = Card> + 'a {
    s.lines().map(|line| Card::from(line))
}

fn count_cards(s: &str) -> u32 {
    let cards = parse_cards(s).collect::<Vec<Card>>();
    let mut card_bins: HashMap<u32, usize> = HashMap::new();
    for card in cards.iter() {
        card_bins.entry(card.number).or_insert(1);
    }

    for card in cards.iter() {
        let match_count = card.match_count();

        if match_count > 0 {
            let cnt = card_bins
                .get(&card.number)
                .map(|c| *c)
                .unwrap_or(0 as usize);

            let start_idx: usize = card.number as usize;
            for cd in cards[start_idx..(start_idx + card.match_count() as usize)].iter() {
                card_bins
                    .entry(cd.number)
                    .and_modify(|e| *e += cnt)
                    .or_insert(1 as usize);
            }
        }
    }

    card_bins
        .into_iter()
        .fold(0 as usize, |acc, (_, val)| acc + val) as u32
}

#[derive(Clone, Debug, PartialEq)]
struct Card {
    number: u32,
    winners: HashSet<u32>,
    actual: HashSet<u32>,
}

impl Card {
    fn points(&self) -> u32 {
        let num_matches = self.winners.intersection(&self.actual).count();
        if num_matches == 0 {
            0
        } else {
            (2 as u32).pow(num_matches as u32 - 1)
        }
    }

    fn match_count(&self) -> u32 {
        self.winners.intersection(&self.actual).count() as u32
    }
}

impl From<&str> for Card {
    fn from(i: &str) -> Self {
        let mut iter_1 = i.split(':');
        let part1 = iter_1.next().unwrap();
        let part2 = iter_1.next().unwrap();

        let mut iter_2 = part2.split('|');
        let winners = iter_2.next().unwrap().trim();
        let actual = iter_2.next().unwrap().trim();

        let number = part1[4..].trim().parse::<u32>().unwrap();

        Self {
            number,
            winners: winners
                .split(' ')
                .filter(|v| v.len() > 0)
                .map(|v| v.trim().parse::<u32>().unwrap())
                .collect::<HashSet<u32>>(),
            actual: actual
                .split(' ')
                .filter(|v| v.len() > 0)
                .map(|v| v.parse::<u32>().unwrap())
                .collect::<HashSet<u32>>(),
        }
    }
}

#[cfg(test)]
mod test {
    use super::*;

    const INPUT: &str = "Card 1: 41 48 83 86 17 | 83 86  6 31 17  9 48 53
Card 2: 13 32 20 16 61 | 61 30 68 82 17 32 24 19
Card 3:  1 21 53 59 44 | 69 82 63 72 16 21 14  1
Card 4: 41 92 73 84 69 | 59 84 76 51 58  5 54 83
Card 5: 87 83 26 28 32 | 88 30 70 12 93 22 82 36
Card 6: 31 18 13 56 72 | 74 77 10 23 35 67 36 11";

    #[test]
    fn it_parses_a_card() {
        let card = Card::from("Card 1: 41 48 83 86 17 | 83 86  6 31 17  9 48 53");
        assert_eq!(
            card,
            Card {
                number: 1,
                winners: vec![41, 48, 83, 86, 17]
                    .into_iter()
                    .collect::<HashSet<u32>>(),
                actual: vec![83, 86, 6, 31, 17, 9, 48, 53]
                    .into_iter()
                    .collect::<HashSet<u32>>(),
            }
        );
        assert_eq!(card.points(), 8);
    }

    #[test]
    fn it_scores_cards() {
        assert_eq!(
            parse_cards(INPUT)
                .map(|card| card.points())
                .collect::<Vec<u32>>(),
            vec![8, 2, 2, 1, 0, 0]
        );
    }

    #[test]
    fn it_counts_winning_scratchcards() {
        assert_eq!(count_cards(INPUT), 30);
    }
}
