use pariter::{scope, IteratorExt as _};
use std::collections::HashMap;

const INPUT: &str = include_str!("../data/day5.txt");

pub fn day5a() -> String {
    let maps = parse_maps(INPUT.lines());
    let seeds = vec![
        2880930400, 17599561, 549922357, 200746426, 1378552684, 43534336, 155057073, 56546377,
        824205101, 378503603, 1678376802, 130912435, 2685513694, 137778160, 2492361384, 188575752,
        3139914842, 1092214826, 2989476473, 58874625,
    ];

    let minimum = scope(|scope| {
        seeds
            .into_iter()
            .parallel_map_scoped(scope, |seed| seed_to_location(seed, &maps))
            .fold(i64::MAX, |acc, val| std::cmp::min(acc, val))
    })
    .unwrap();

    format!("{}", minimum)
}

pub fn day5b() -> String {
    let maps = parse_maps(INPUT.lines());
    let seed_ranges: Vec<(i64, i64)> = vec![
        (2880930400, 17599561),
        (549922357, 200746426),
        (1378552684, 43534336),
        (155057073, 56546377),
        (824205101, 378503603),
        (1678376802, 130912435),
        (2685513694, 137778160),
        (2492361384, 188575752),
        (3139914842, 1092214826),
        (2989476473, 58874625),
    ];
    let minimum = seed_ranges
        .into_iter()
        .map(|(start, len)| {
            let start_time = std::time::SystemTime::now();
            println!("scanning range: {} {}", start, len);
            let min = minimum_of_range((start, len), &maps).unwrap();
            let end_time = std::time::SystemTime::now();
            println!(
                "minimum found for ({}, {}), {} ({} seconds)",
                start,
                len,
                min,
                end_time.duration_since(start_time).unwrap().as_secs()
            );
            min
        })
        .fold(i64::MAX, std::cmp::min);
    format!("{}", minimum)
}

fn minimum_of_range(
    (start, len): (i64, i64),
    maps: &HashMap<MapType, Vec<Map>>,
) -> Result<i64, Box<(dyn std::any::Any + Send + 'static)>> {
    scope(|scope| {
        (start..start + len)
            .parallel_map_scoped(scope, |seed| seed_to_location(seed, maps))
            .fold(i64::MAX, |min, loc| std::cmp::min(min, loc))
    })
}

fn seed_to_location(seed: i64, maps: &HashMap<MapType, Vec<Map>>) -> i64 {
    let mut current = seed;
    for ty in MAP_TYPES {
        current = run_maps(current, maps.get(&ty).unwrap());
    }
    current
}

fn run_maps(input: i64, maps: &Vec<Map>) -> i64 {
    for map in maps {
        if let Some(dest) = map.lookup(input) {
            return dest;
        }
    }
    input
}

#[derive(Clone, Copy, Debug, PartialEq, Eq, Hash)]
enum MapType {
    SeedToSoil,
    SoilToFertilizer,
    FertilizerToWater,
    WaterToLight,
    LightToTemperature,
    TemperatureToHumidity,
    HumidityToLocation,
}

const MAP_TYPES: [MapType; 7] = [
    MapType::SeedToSoil,
    MapType::SoilToFertilizer,
    MapType::FertilizerToWater,
    MapType::WaterToLight,
    MapType::LightToTemperature,
    MapType::TemperatureToHumidity,
    MapType::HumidityToLocation,
];

#[derive(Clone, Debug, PartialEq)]
struct Map {
    map_type: MapType,
    source_start: i64,
    dest_start: i64,
    length: i64,
}

impl Map {
    fn lookup(&self, val: i64) -> Option<i64> {
        let offset: i64 = (val as i64) - (self.source_start as i64);
        if offset >= 0 && offset < self.length as i64 {
            Some((self.dest_start as i64 + offset) as i64)
        } else {
            None
        }
    }
}

fn parse_maps<'a>(lines: impl IntoIterator<Item = &'a str>) -> HashMap<MapType, Vec<Map>> {
    let mut ty = None;
    let mut maps: HashMap<MapType, Vec<Map>> = HashMap::new();
    for line in lines {
        if line.starts_with("seed-to-soil") {
            ty = Some(MapType::SeedToSoil);
        } else if line.starts_with("soil-to-fertilizer") {
            ty = Some(MapType::SoilToFertilizer);
        } else if line.starts_with("fertilizer-to-water") {
            ty = Some(MapType::FertilizerToWater);
        } else if line.starts_with("water-to-light") {
            ty = Some(MapType::WaterToLight);
        } else if line.starts_with("light-to-temperature") {
            ty = Some(MapType::LightToTemperature);
        } else if line.starts_with("temperature-to-humidity") {
            ty = Some(MapType::TemperatureToHumidity);
        } else if line.starts_with("humidity-to-location") {
            ty = Some(MapType::HumidityToLocation);
        } else if line.is_empty() {
        } else {
            match ty {
                Some(ty) => {
                    let map = parse_map(line, ty);
                    maps.entry(ty)
                        .and_modify(|map_list| map_list.push(map.clone()))
                        .or_insert(vec![map]);
                }
                None => {}
            }
        }
    }

    maps
}

fn parse_map(line: &str, ty: MapType) -> Map {
    let mut parts = line.split(' ');
    Map {
        map_type: ty,
        dest_start: parts.next().unwrap().parse::<i64>().unwrap(),
        source_start: parts.next().unwrap().parse::<i64>().unwrap(),
        length: parts.next().unwrap().parse::<i64>().unwrap(),
    }
}

#[cfg(test)]
mod test {
    use super::*;

    const INPUT: &str = "seeds: 79 14 55 13

seed-to-soil map:
50 98 2
52 50 48

soil-to-fertilizer map:
0 15 37
37 52 2
39 0 15

fertilizer-to-water map:
49 53 8
0 11 42
42 0 7
57 7 4

water-to-light map:
88 18 7
18 25 70

light-to-temperature map:
45 77 23
81 45 19
68 64 13

temperature-to-humidity map:
0 69 1
1 0 69

humidity-to-location map:
60 56 37
56 93 4";

    #[test]
    fn it_parses_a_map() {
        let map = parse_map("50 98 2", MapType::SeedToSoil);
        assert_eq!(
            map,
            Map {
                map_type: MapType::SeedToSoil,
                source_start: 98,
                dest_start: 50,
                length: 2
            }
        );
        assert_eq!(map.lookup(49), None);
        assert_eq!(map.lookup(51), None);
        assert_eq!(map.lookup(98), Some(50));
        assert_eq!(map.lookup(99), Some(51));
        assert_eq!(map.lookup(100), None);

        let map = parse_map("52 50 48", MapType::SeedToSoil);
        assert_eq!(map.lookup(79), Some(81));
    }

    #[test]
    fn it_parses_maps() {
        let maps = parse_maps(INPUT.lines());

        assert_eq!(maps.get(&MapType::SeedToSoil).map(|m| m.len()).unwrap(), 2);
        assert_eq!(
            maps.get(&MapType::SoilToFertilizer)
                .map(|m| m.len())
                .unwrap(),
            3
        );
        assert_eq!(
            maps.get(&MapType::FertilizerToWater)
                .map(|m| m.len())
                .unwrap(),
            4
        );
        assert_eq!(
            maps.get(&MapType::WaterToLight).map(|m| m.len()).unwrap(),
            2
        );
        assert_eq!(
            maps.get(&MapType::LightToTemperature)
                .map(|m| m.len())
                .unwrap(),
            3
        );
        assert_eq!(
            maps.get(&MapType::TemperatureToHumidity)
                .map(|m| m.len())
                .unwrap(),
            2
        );
        assert_eq!(
            maps.get(&MapType::HumidityToLocation)
                .map(|m| m.len())
                .unwrap(),
            2
        );
    }

    #[test]
    fn it_maps_seed_to_location() {
        let maps = parse_maps(INPUT.lines());
        assert_eq!(seed_to_location(79, &maps), 82);
        assert_eq!(seed_to_location(14, &maps), 43);
        assert_eq!(seed_to_location(55, &maps), 86);
        assert_eq!(seed_to_location(13, &maps), 35);
    }

    #[test]
    fn it_maps_by_seed_range() {
        let seed_ranges: Vec<(i64, i64)> = vec![(79, 14), (55, 13)];
        let maps = parse_maps(INPUT.lines());

        assert_eq!(minimum_of_range(seed_ranges[0], &maps).unwrap(), 46);
    }
}
