use nom::{
    bytes::complete::{tag, take_while},
    character::{
        complete::{alpha1, alphanumeric1},
        is_alphabetic,
    },
    IResult, Parser,
};
use std::collections::HashMap;

const INPUT: &str = include_str!("../data/day8.txt");

pub fn day8a() -> String {
    format!("{}", solve_steps(INPUT))
}

pub fn day8b() -> String {
    format!("{}", parallel_solve_steps(INPUT))
}

fn solve_steps(input: &str) -> u32 {
    let (path, map) = parse_data(input);
    solve_path(&path, "AAA".to_string(), |s| s == "ZZZ", &map)
}

fn solve_path<F>(path: &str, start: String, end_criteria: F, map: &Map) -> u32
where
    F: Fn(&str) -> bool,
{
    let mut current = start.clone();
    let mut cnt = 0;
    while !end_criteria(current.as_ref()) {
        for step in path.chars() {
            let node = map.get(&current).unwrap();
            match step {
                'L' => {
                    current = node.left.clone();
                }
                'R' => {
                    current = node.right.clone();
                }
                _ => unreachable!(),
            }
            cnt += 1;
        }
    }
    cnt
}

fn parallel_solve_steps(input: &str) -> u32 {
    let (path, map) = parse_data(input);
    let mut cnt = 0;
    let mut starting_addrs = map
        .keys()
        .filter(|k| k.ends_with('A'))
        .cloned()
        .collect::<Vec<String>>();

    for addr in starting_addrs {
        let cnt = solve_path(&path, addr.to_owned(), |s| s.ends_with('Z'), &map);
        println!("{}: {}", addr, cnt);
    }

    0
}

fn is_finished(addrs: &Vec<String>) -> bool {
    addrs.iter().all(|addr| addr.ends_with('Z'))
}

type Map = HashMap<String, Node>;

#[derive(Clone, Debug, PartialEq)]
struct Node {
    left: String,
    right: String,
}

fn parse_data(input: &str) -> (String, Map) {
    let mut lines = input.lines();
    let path = lines.next().unwrap();
    let _ = lines.next();
    let map = parse_map(lines.collect::<Vec<&str>>().join("\n").as_ref());
    (path.to_owned(), map)
}

fn parse_map(input: &str) -> Map {
    input
        .lines()
        .map(|line| {
            let (_, (addr, node)) = parse_line(line).unwrap();
            (addr, node)
        })
        .fold(HashMap::new(), |mut acc, (addr, node)| {
            acc.insert(addr, node);
            acc
        })
}

fn parse_line(input: &str) -> IResult<&str, (String, Node)> {
    let (input, addr) = parse_addr(input)?;
    let (input, _) = tag(" = (")(input)?;
    let (input, left) = parse_addr(input)?;
    let (input, _) = tag(", ")(input)?;
    let (input, right) = parse_addr(input)?;
    Ok((input, (addr, Node { left, right })))
}

fn parse_addr(input: &str) -> IResult<&str, String> {
    let (input, s) = alphanumeric1(input)?;
    Ok((input, s.to_owned()))
}

#[cfg(test)]
mod test {
    use super::*;

    const INPUT1: &str = "RL

AAA = (BBB, CCC)
BBB = (DDD, EEE)
CCC = (ZZZ, GGG)
DDD = (DDD, DDD)
EEE = (EEE, EEE)
GGG = (GGG, GGG)
ZZZ = (ZZZ, ZZZ)";

    const INPUT2: &str = "LLR

AAA = (BBB, BBB)
BBB = (AAA, ZZZ)
ZZZ = (ZZZ, ZZZ)";

    const INPUT3: &str = "LR

11A = (11B, XXX)
11B = (XXX, 11Z)
11Z = (11B, XXX)
22A = (22B, XXX)
22B = (22C, 22C)
22C = (22Z, 22Z)
22Z = (22B, 22B)
XXX = (XXX, XXX)";

    #[test]
    fn solves_steps() {
        assert_eq!(solve_steps(INPUT1), 2);
        assert_eq!(solve_steps(INPUT2), 6);
    }

    #[test]
    fn solves_in_parallel() {
        assert_eq!(parallel_solve_steps(INPUT3), 6);
    }

    #[test]
    fn parses_data() {
        let (path, map) = parse_data(INPUT2);
        assert_eq!(path, "LLR");
        println!("{:?}", map);
        assert_eq!(
            *map.get("AAA").unwrap(),
            Node {
                left: "BBB".to_owned(),
                right: "BBB".to_owned(),
            }
        );
    }
}
